# (c) 2022 DTU Wind Energy
"""Geostrophic wind climate module

A geostrophic wind climate is somehwat similar as a binned wind climate, but
contains 6 variables:
geo_wv_freq is the geostrophic wind speed frequency binned by wind speed
and wind direction, where the whole histogram sums to 1 (note
that in a binned wind climate the frequency in one sector sums to one).
The geostrophic turning is the turning of the wind in the boundary layer
for each bin given by the geostrophic drag law.

A valid geostrophic wind climate therefore has dimensions ``sector`` and ``wsbin``
and variables ``geo_wv_freq``, ``geo_turn``, ``z0meso``, ``slfmeso``, ``displ``,
 ``flow_sep_height``. Also it must have a valid spatial structure.
"""

__all__ = ["validate_geowc", "is_geowc"]

from windkit.xarray_structures._validate import (
    _create_is_obj_function,
    _create_validation_wrapper_factory,
    _create_validator,
)

DATA_VAR_DICT_GEOWC = {
    "geo_wv_freq": ["wsbin", "sector"],
    "geo_turn": ["wsbin", "sector"],
    "z0meso": ["sector"],
    "slfmeso": ["sector"],
    "displ": ["sector"],
    "flow_sep_height": ["sector"],
}

REQ_DIMS_GEOWC = ["wsbin", "sector"]

REQ_COORDS_GEOWC = [
    "sector",
    "wsbin",
]

validate_geowc = _create_validator(
    variables=DATA_VAR_DICT_GEOWC,
    dims=REQ_DIMS_GEOWC,
    coords=REQ_COORDS_GEOWC,
    extra_checks=[],
)

_validate_geowc_wrapper_factory = _create_validation_wrapper_factory(validate_geowc)

is_geowc = _create_is_obj_function(validate_geowc)


def _mean_ws_moment(geowc, moment=1, **kwargs):
    """Calculates mean wind speed moment from geostrophic wind climate dataset."""
    raise NotImplementedError("This function is not yet implemented.")


def _ws_cdf(geowc, **kwargs):
    """Calculates wind speed cumulative distribution function from geostrophic wind climate dataset."""
    raise NotImplementedError("This function is not yet implemented.")


def _ws_freq_gt_mean(geowc, **kwargs):
    """Calculates wind speed frequency greater than mean from geostrophic wind climate dataset."""
    raise NotImplementedError("This function is not yet implemented.")


def _mean_wind_speed(geowc, **kwargs):
    """Calculates mean wind speed from geostrophic wind climate dataset."""
    raise NotImplementedError("This function is not yet implemented.")


def _mean_power_density(geowc, **kwargs):
    """Calculates mean power density from geostrophic wind climate dataset."""
    raise NotImplementedError("This function is not yet implemented.")
